<?php

namespace Myriad\EventDrivenService\Services;

use Carbon\Carbon;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\Auth;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Validator;
use Illuminate\Validation\ValidationException;
use Myriad\EventDrivenService\Http\Utilities\EventBuilder;
use Myriad\EventDrivenService\Http\Utilities\EventHelper;
use Myriad\EventDrivenService\Http\Utilities\Pagination;
use Myriad\EventDrivenService\Interfaces\IEventRepository;
use Myriad\EventDrivenService\Interfaces\IFailedEventRepository;
use Myriad\EventDrivenService\Models\EventHadAccess;
use Myriad\EventDrivenService\Models\EventModel;
use Myriad\EventDrivenService\Rules\UuidRule;
use Myriad\EventDrivenService\Traits\EventExceptionHandler;
use Ramsey\Uuid\Uuid;
use StableSort;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;

class EventService {
  use EventExceptionHandler;

  private $eventRepository;
  private $failedEventRepository;
  private $modelCache = [];

  private $customCommands = [
    'batch' => 'customCommandBatch',
  ];

  public function __construct(
    IEventRepository $eventRepository,
    IFailedEventRepository $failedEventRepository
  ) {
    $this->eventRepository = $eventRepository;
    $this->failedEventRepository = $failedEventRepository;
  }

  public function getNew($uri, $lastUpdate, $pageSize) {
    $model = EventHelper::getModelByUri($uri);
    $modelType = EventHelper::getCommandForModel($model);

    $trackAccessChanges =
      property_exists($model, 'trackAccessChanges') && $model->trackAccessChanges;
    $needSync = $model->needsFullSync($lastUpdate);
    $recordsFound = $model->getUpdated($lastUpdate, $pageSize);

    $currentData = [];
    $toMarkAsDeleted = [];
    $toMarkAsAccessible = [];
    foreach ($recordsFound as $record) {
      $event = EventBuilder::make($modelType, $record);
      if ($trackAccessChanges) {
        if ($event['command'] == $modelType . ':delete') {
          $toMarkAsDeleted[] = $record->uuid;
        } else {
          $toMarkAsAccessible[] = $record->uuid;
        }
      }
      $currentData[] = $event;
    }

    if ($trackAccessChanges) {
      $existingUUIDs = EventHadAccess::query()
        ->select('model_uuid')
        ->whereIn('model_uuid', array_merge($toMarkAsDeleted, $toMarkAsAccessible))
        ->where('user_uuid', Auth::id())
        ->where('model', $modelType)
        ->get()
        ->pluck('model_uuid')
        ->toArray();

      // mark accessible models as accessed
      $accessibleData = ['last_requested_at' => Carbon::now(), 'lost_at' => null];
      EventHadAccess::query()
        ->whereIn('model_uuid', $toMarkAsAccessible)
        ->where('user_uuid', Auth::id())
        ->where('model', $modelType)
        ->update($accessibleData);

      // mark deleted models as lost access
      $deletedData = ['lost_at' => Carbon::now()];
      EventHadAccess::query()
        ->whereIn('model_uuid', $toMarkAsDeleted)
        ->where('user_uuid', Auth::id())
        ->where('model', $modelType)
        ->update($deletedData);

      // insert new models
      $commonInsertData = [
        'user_uuid' => Auth::id(),
        'model' => $modelType,
        'last_requested_at' => Carbon::now(),
        'lost_at' => null,
      ];
      $toInsert = [];
      foreach (array_diff($toMarkAsAccessible, $existingUUIDs) as $uuid) {
        $toInsert[] = array_merge($commonInsertData, $accessibleData, ['model_uuid' => $uuid]);
      }
      foreach (array_diff($toMarkAsDeleted, $existingUUIDs) as $uuid) {
        $toInsert[] = array_merge($commonInsertData, $deletedData, ['model_uuid' => $uuid]);
      }
      EventHadAccess::query()->insert($toInsert);
    }

    // build response
    $meta = Pagination::build($recordsFound);
    $meta['full_sync_required'] = $needSync;

    return [
      'data' => collect($currentData),
      'meta' => $meta,
    ];
  }

  public function handleBulkEvents(array $events) {
    try {
      DB::beginTransaction();
      list($errors) = $this->handlePost($events, false, true);
      if (count($errors) > 0) {
        DB::rollBack();
        $this->failedEventRepository->createAll($errors);
        return response()->json(
          [
            'errors' => $errors
              ->groupBy([
                function ($value, $key) {
                  return explode(':', $value['request']['command'])[0];
                },
                function ($value, $key) {
                  return explode(':', $value['request']['command'])[1];
                },
                function ($value, $key) {
                  return explode(':', $value['request']['command'])[1] !== 'update'
                    ? 'all'
                    : @$value['request']['name'];
                },
              ])
              ->map(function ($commandLevel) {
                return $commandLevel->map(function ($fieldLevel, $command) {
                  if ($command !== 'update') {
                    return $fieldLevel->first()->first()['error'];
                  }
                  return $fieldLevel->map(function ($updateLevel) {
                    return $updateLevel->first()['error'];
                  });
                });
              })
              ->toArray(),
          ],
          400
        );
      }
      DB::commit();
    } catch (\Exception $e) {
      DB::rollBack();
      throw $e;
    }
    return response()->json([], 201);
  }

  public function handlePost(
    array $events,
    $processUpdateResponse,
    $isBulk = false,
    $saveFailures = true,
    $additionalMeta = []
  ) {
    list($valid, $invalid) = collect($events)->partition(function ($value, $key) {
      return isset($value['occurred_at']) &&
        $this->isDateStringValid($value['occurred_at']) &&
        isset($value['command']) &&
        ((isset($value['model_uuid']) && Uuid::isValid($value['model_uuid'])) ||
          array_key_exists($value['command'], $this->customCommands));
    });

    $validArray = $valid->toArray();
    StableSort::usort($validArray, function ($a, $b) {
      return new \DateTime($a['occurred_at']) <=> new \DateTime($b['occurred_at']);
    });
    $valid = collect($validArray);

    $errors = $invalid->map(function ($value, $key) {
      return $this->formatError($value, __('event-driven-service::events.invalid'), true);
    });

    $successfulEvents = [];
    $updateResponse = [];
    $lastIndices = $valid
      ->whereNotIn('command', array_keys($this->customCommands))
      ->groupBy(
        [
          function ($key) {
            list($model) = explode(':', $key['command']);
            return $model;
          },
          'model_uuid',
        ],
        true
      )
      ->map(function ($v) {
        return $v->map(function ($v) {
          return $v->keys()->last();
        });
      });

    foreach ($valid as $i => $event) {
      if (array_key_exists($event['command'], $this->customCommands)) {
        [$newErrors, $newUpdateResponse] = call_user_func_array(
          [$this, $this->customCommands[$event['command']]],
          [$event, $processUpdateResponse, $isBulk]
        );
        if ($newErrors) {
          $errors = $errors->merge($newErrors);
        }
        if ($newUpdateResponse) {
          $updateResponse = array_merge($updateResponse, $newUpdateResponse);
        }
        continue;
      }

      list($modelType, $command) = explode(':', $event['command']);
      $event['model'] = $modelType;
      $event['operation'] = $command;
      $uuid = $event['model_uuid'];

      $currentModel = $this->modelCache[$modelType][$uuid] ?? null;

      if (!$currentModel) {
        $model = EventHelper::getModelByType($modelType);
        $currentModel = $model
          ->withAll()
          ->where($model->modelKeyField, $uuid)
          ->first();
        $created = false;
        if (!$currentModel) {
          $created = true;
          $modelData = [
            $model->modelKeyField => $uuid,
          ];
          if ($model->modelKeyField != 'uuid') {
            $modelData['uuid'] = $model->generateUuid();
          }
          $currentModel = $model->make($modelData)->refresh();
        }
      }

      $this->modelCache[$modelType][$uuid] = $currentModel;

      if ($currentModel->trashed()) {
        $errors = $errors->merge([
          'request' => $event,
          'error' => __('event-driven-service::events.deleted'),
        ]);
        continue;
      }

      $event['is_bulk'] = $isBulk;
      $event['last_event_in_batch'] = $i === $lastIndices->get($modelType)->get($uuid);
      try {
        Validator::make($event, [
          'operation' => 'required|in:' . implode(',', $currentModel->getCommands()),
        ])->validate();

        if ($event['operation'] == 'create') {
          if (!$created && !$currentModel->canRecreate()) {
            throw new BadRequestHttpException(__('event-driven-service::events.created'));
          }
          if (is_string($event['value'])) {
            $event['value'] = json_decode($event['value'], true);
          }
        }

        if (method_exists($this, $event['operation'])) {
          $response = call_user_func_array(
            [$this, $event['operation']],
            [$currentModel, $event, $additionalMeta]
          );
        } else {
          $response = call_user_func_array(
            [$currentModel, $event['operation']],
            [$event, $additionalMeta]
          );
        }

        if ($processUpdateResponse) {
          $updateResponse = array_merge($updateResponse, $currentModel->getUpdatesResponse());
        }

        if ($response) {
          $errors[] = $response;
        } else {
          $successfulEvents[] = $event;
        }
      } catch (ValidationException $e) {
        $this->logError($e);
        $errors[] = $this->handleValidationError($event, $e);
        continue;
      } catch (\Exception $e) {
        $this->logError($e);
        $errors[] = $this->formatError($event, $e->getMessage());
        continue;
      }
    }
    if ($saveFailures) {
      $this->failedEventRepository->createAll($errors);
    }
    return [$errors, $updateResponse, $successfulEvents];
  }

  public function create(EventModel $model, $event, $meta) {
    try {
      Validator::make($event, [
        'value' => 'required',
      ])->validate();
    } catch (ValidationException $e) {
      throw $e;
    }

    $event['value']['uuid'] = $model->uuid;

    try {
      DB::beginTransaction();
      $this->eventRepository->create($event);
      $model->createModel($event, $meta);
      DB::commit();
    } catch (\Exception $e) {
      DB::rollBack();
      throw $e;
    }
  }

  public function update(EventModel $model, $event, $meta) {
    try {
      Validator::make($event, [
        'name' => 'required|string',
        'value' => 'nullable',
        'version' => ['nullable', new UuidRule()],
        'version_new' => ['required', new UuidRule()],
      ])->validate();
    } catch (ValidationException $e) {
      throw $e;
    }

    $nonColumnField = in_array($event['name'], $model->getNonColumnFields());

    if (!in_array($event['name'], $model->getFillable()) && !$nonColumnField) {
      throw new BadRequestHttpException(__('event-driven-service::events.update.not_found'));
    }

    if (!in_array($event['name'] . '_version', $model->getFillable()) && !$nonColumnField) {
      throw new BadRequestHttpException(__('event-driven-service::events.update.offline'));
    }

    if (!$model->isCurrent($event['name'], $event['version']) && !$nonColumnField) {
      throw new BadRequestHttpException(__('event-driven-service::events.update.version'));
    }

    try {
      DB::beginTransaction();
      $this->eventRepository->create($event);
      $model->updateModel($event, $meta);
      DB::commit();
    } catch (\Exception $e) {
      DB::rollBack();
      throw $e;
    }
  }

  public function delete(EventModel $model, $event, $meta) {
    try {
      DB::beginTransaction();
      $this->eventRepository->create($event);
      $model->deleteModel($event, $meta);
      DB::commit();
    } catch (\Exception $e) {
      DB::rollBack();
      throw $e;
    }
  }

  public function customCommandBatch($event, $processUpdateResponse) {
    if (!array_key_exists('events', $event)) {
      return [[$this->formatError($event, 'Batch command requires events array.')], []];
    }
    \DB::beginTransaction();
    try {
      [$newErrors, $newUpdateResponse, $successfulEvents] = $this->handlePost(
        $event['events'],
        $processUpdateResponse,
        true,
        false,
        ['transaction' => $event['events']]
      );
      if ($newErrors->count() > 0) {
        $newErrors = $newErrors->merge(
          collect($successfulEvents)->map(function ($e) {
            return $this->formatError($e, 'Failure in transaction.');
          })
        );
        \DB::rollBack();
      } else {
        \DB::commit();
      }
      return [$newErrors, $newUpdateResponse];
    } catch (\Exception $e) {
      \DB::rollBack();
      throw $e;
    }
  }

  private function isDateStringValid($date) {
    try {
      Carbon::createFromFormat(EventHelper::OCCURRED_AT_FORMAT, $date);
      return true;
    } catch (\Exception $e) {
      return false;
    }
  }
}
